module Main exposing (..)

import Browser
import Html exposing (..)
import Html.Attributes exposing (..)
import Html.Events exposing (..)
import Svg exposing (svg)
import Svg.Attributes
import Character exposing (Character)
import Stats exposing (Stat)
import BoundedInt exposing (BoundedInt)
import Stats exposing (Category)
import Creation exposing (CreationPoints)
import OrderedCategories as OCat exposing (OCat, CategoriesType)
import Http
import Json.Decode

main : Program () Model Msg
main = 
    Browser.document
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        }

init : () -> ( Model, Cmd Msg )
init _ =
    (   { character = Character.new
        , modal = Nothing
        , modalValue = ""
        , dropdownActive = False
        }
    , Cmd.none
    )

type alias Model = 
    { character : Character
    , modal : Maybe ModalType
    , modalValue : String
    , dropdownActive : Bool
    }

type Msg 
    = OpenModal ModalType
    | CloseModal
    | ChangeModalValue String
    | SaveModalChange ModalType
    | ChangeStat Stat
    | ChangePriority CategoriesType Category Int
    | SetDropDownActive Bool
    | SetAffinity Stat
    | ImportCharacter
    | JsonContent (Result Http.Error Character )

type ModalType
    = Name

update : Msg -> Model -> ( Model, Cmd Msg )
update msg model = 
    case msg of
        OpenModal modalType -> 
            (   { model 
                | modal = ( Just modalType ) 
                , modalValue = modalValue modalType model.character
                , dropdownActive = False
                }
            , Cmd.none
            )
        CloseModal ->
            ( { model | modal = Nothing }
            , Cmd.none
            )
        ChangeModalValue value ->
            ( { model | modalValue = value }
            , Cmd.none
            )
        SaveModalChange modalType ->
            case modalType of
                Name ->
                    (   { model 
                        | character = Character.changeName model.character model.modalValue 
                        , modal = Nothing
                        }
                    , Cmd.none
                    )
        ChangeStat stat ->
            (   { model 
                | character = Character.changeStat model.character stat
                , dropdownActive = False
                }
            , Cmd.none
            )
        ChangePriority catType category value ->
            (   { model 
                | character = Character.changeOrder model.character catType category value
                , dropdownActive = False
                }
            , Cmd.none
            )
        SetDropDownActive value ->
            (   { model
                | dropdownActive = value
                }
            , Cmd.none
            )
        SetAffinity stat ->
            (   { model
                | character = Character.setAffinity model.character stat
                }
            , Cmd.none
            )
        ImportCharacter ->
            ( model
            , Http.get
                { url = "./template.json"
                , expect = Http.expectJson JsonContent jsonDecoder
                }
            )
        JsonContent content ->
            (   { model
                | character = case content of
                    Ok character -> character
                    Err _ -> model.character
                }
            , Cmd.none
            )

view : Model -> Browser.Document Msg
view model =
    { title = "Character Sheet"
    , body =
        [ div 
            [ class "container" ] 
            [ viewEditableText Name model.character.name
            , viewAttributes model.character.creationPoints.attributes model.character.stats
            , viewAbilities model.character.creationPoints.abilities model.character.stats
            , viewBackgrounds model.dropdownActive model.character.creationPoints ( Stats.backgroundsInList model.character.stats )
            , viewSpheres model.character.creationPoints ( Stats.spheresInList model.character.stats )
            , viewImport
            , viewModal model
            ]
        ]
    }

viewEditableText : ModalType -> String -> Html Msg
viewEditableText modalType content =
    div []
        [ span [ class "icon-text" ]
            [ span [] [ text content ]
            , span 
                [ class "icon" 
                ] 
                [ i 
                    [ class "fas fa-pen"
                    , onClick ( OpenModal modalType ) 
                    ] []
                ]
            ]
        ]

viewAttributes : OCat -> List Stat -> Html Msg
viewAttributes oCat attributes =
    div [ class "has-text-centered" ]
        [ p [] [ text "Attributes" ]
        , div [ class "columns is-centered is-mobile" ]
            [ viewAttributeCol oCat "Physical" Stats.Physical attributes
            , viewAttributeCol oCat "Social" Stats.Social attributes
            , viewAttributeCol oCat "Mental" Stats.Mental attributes
            ]
        ]

viewAttributeCol : OCat -> String -> Category -> List Stat -> Html Msg
viewAttributeCol oCat name category attributes =
    let 
        categoryAttributes : List Stat
        categoryAttributes = Stats.attributesOfCategoryInList category attributes

        index : Int
        index = OCat.indexOf oCat.order category
        basePoints : Int
        basePoints = OCat.categoryValue oCat category
        points : Int
        points = basePoints - ( Stats.count categoryAttributes ) + ( List.length categoryAttributes )
    in
        div [ class "column is-one-fifth" ]
            (   [ p [] [ text name ]
                , creationInfo points
                , div [ class "field has-addons is-fullwidth" ]
                    [ p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 0 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 0 )
                            ] [ text "I" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 1 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 1 )
                            ] [ text "II" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 2 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 2 )
                            ] [ text "III" ]
                        ]
                    ]
                ]
                ++ 
                ( List.map viewStat categoryAttributes
                )
            )

viewAbilities : OCat -> List Stat -> Html Msg
viewAbilities oCat abilities =
    div [ class "has-text-centered" ]
        [ p [] [ text "Abilities" ]
        , div [ class "columns is-centered is-desktop" ]
            [ viewAbilityCol oCat "Talents" Stats.Talent abilities
            , viewAbilityCol oCat "Skills" Stats.Skill abilities
            , viewAbilityCol oCat "Knowledges" Stats.Knowledge abilities
            ]
        ]

viewAbilityCol : OCat -> String -> Category -> List Stat -> Html Msg
viewAbilityCol oCat name category abilities =
    let 
        categoryAbilities : List Stat
        categoryAbilities = Stats.abilitiesOfCategoryInList category abilities

        index : Int
        index = OCat.indexOf oCat.order category
        basePoints : Int
        basePoints = OCat.categoryValue oCat category
        points : Int
        points = basePoints - ( Stats.count categoryAbilities )
    in
        div [ class "column" ]
            (   [ p [] [ text name ]
                , creationInfo points
                , div [ class "field has-addons is-desktop" ]
                    [ p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 0 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 0 )
                            ] [ text "I" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 1 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 1 )
                            ] [ text "II" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 2 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 2 )
                            ] [ text "III" ]
                        ]
                    ]
                , div [ class "fixed-grid has-6-cols" ]
                    [ div [ class "grid" ]
                        ( List.map ( viewStatInGrid 3 ( List.length categoryAbilities ) ) ( List.indexedMap Tuple.pair categoryAbilities )
                        )
                    ]
                ]
            )

viewStatInGrid : Int -> Int -> ( Int, Stat ) -> Html Msg
viewStatInGrid cols totalCount indexedStat =
    let
        index : Int
        index = Tuple.first indexedStat
        stat : Stat
        stat = Tuple.second indexedStat

        offset : Int
        offset = Basics.remainderBy cols totalCount
        dif : Int
        dif = totalCount - index

        offsetString : String
        offsetString = 
            if dif == offset
            then " is-col-start-" ++ ( String.fromInt ( 1 + cols - dif ) )
            else ""
    in
        div [ class ( "cell is-col-span-2" ++ offsetString ) ]
            [ div [ class "notification" ]
                [ p [ class "block" ] [ text stat.name ]
                , p [ class "block" ] [ viewDots stat ]
                ]
            ]

viewBackgrounds : Bool -> CreationPoints -> List Stat -> Html Msg
viewBackgrounds dropdownActive cp backgrounds =
    let
        basePoints : Int
        basePoints = cp.backgrounds
        points : Int
        points = basePoints - ( Stats.count backgrounds )

        separatedBackgrounds : ( List Stat, List Stat )
        separatedBackgrounds = Stats.splitZero backgrounds
    in
        div [ class "has-text-centered" ]
            (   [ p [] [ text "Backgrounds" ]
                , div [ class "columns is-centered is-mobile" ]
                    [ div [ class "column is-two-fifths" ]
                        (   [ creationInfo points
                            , viewOtherBackgrounds dropdownActive ( Tuple.second separatedBackgrounds )
                            ]
                            ++
                            ( List.map viewStatHorizontal ( Tuple.first separatedBackgrounds ) )
                        )
                    ]
                ]
            )

viewOtherBackgrounds : Bool -> List Stat -> Html Msg
viewOtherBackgrounds dropdownActive backgrounds =
    div [ class ( "block dropdown is-fullwidth" ++ if dropdownActive then " is-active" else "" ) ]
        [ div [ class "dropdown-trigger" ]
            [ div 
                [ class "button" 
                , ariaHaspopup "true"
                , ariaControls "dropdown-menu" 
                , onClick ( SetDropDownActive ( not dropdownActive ) )
                ]
                [ span [] [ text "Add a Background" ]
                , span [ class "icon is-small" ]
                    [ i 
                        [ class "fas fa-angle-down"
                        , ariaHidden "true"
                        ] [] 
                    ]
                ]
            ]
        , div 
            [ class "dropdown-menu"
            , id "dropdown-menu"
            , ariaRole "menu"
            ]
            [ div [ class "dropdown-content has-text-left" ] 
                ( List.map viewStatDropdown backgrounds )
            ]
        ]

viewStatHorizontal : Stat -> Html Msg
viewStatHorizontal stat =
    div [ class "notification is-narrow" ]
        [ div [ class "columns is-centered is-vcentered is-mobile" ]
            [ div [ class "column has-text-right" ] [ text stat.name ]
            , div [ class "column has-text-left" ] [ viewDots stat ]
            ]
        ]

viewStatDropdown : Stat -> Html Msg
viewStatDropdown stat =
    a 
        [ class "dropdown-item" 
        , onClick ( ChangeStat { stat | dots = BoundedInt.changeTo stat.dots 1 } )
        ] [ text stat.name ]
    

viewSpheres : CreationPoints -> List Stat -> Html Msg
viewSpheres cp spheres =
    let
        basePoints : Int
        basePoints = cp.spheres
        points : Int
        points = basePoints - ( Stats.count spheres )
    in
        div [ class "has-text-centered" ]
            [ p [] [ text "Spheres" ]
            , div [ class "columns is-centered is-mobile" ]
                [ div [ class "column is-three-fifths" ]
                    [ creationInfo points
                    , affinityWarning spheres
                    ]
                ]
            , div [ class "columns is-centered is-mobile" ]
                [ div [ class "column is-one-fifth" ] 
                    ( List.map viewSphere ( Stats.everyThirdStat spheres 0 )
                    )
                , div [ class "column is-one-fifth" ] 
                    ( List.map viewSphere ( Stats.everyThirdStat spheres 1 )
                    )
                , div [ class "column is-one-fifth" ] 
                    ( List.map viewSphere ( Stats.everyThirdStat spheres 2 )
                    )
                ]
            ]

viewSphere : Stat -> Html Msg
viewSphere stat =
    div [ class "notification" ]
        [ p 
            [ class "delete is-not-delete icon corner"
            , onClick ( SetAffinity { stat | statType = Stats.Sphere True } )
            ] 
            [ case stat.statType of
                Stats.Sphere affinity ->
                    if affinity 
                    then i [ class "fas fa-star" ] []
                    else i [ class "far fa-star" ] []
                _ -> i [] []
            ]
        , p [ class "block" ] [ text stat.name ]
        , p [ class "block" ] [ viewDots stat ]
        ]

affinityWarning : List Stat -> Html Msg
affinityWarning spheres = 
    let
        hasSetAffinity : Bool
        hasSetAffinity = Stats.hasSetAffinity spheres
    in
        if hasSetAffinity
        then div [] []
        else div [ class "notification is-danger" ] [ text "Affinity Sphere not set" ]

creationInfo : Int -> Html Msg
creationInfo points = 
    let
        notificationColor : String
        notificationColor =
            case compare points 0 of
                GT -> "is-warning"
                EQ -> "is-success"
                LT -> "is-danger"
    in
        div [ class ( "notification " ++ notificationColor ) ]
            [ text ( ( String.fromInt points ) ++ " remaining" )
            ]

viewStat : Stat -> Html Msg
viewStat stat =
    div [ class "notification" ]
        [ p [ class "block" ] [ text stat.name ]
        , p [ class "block" ] [ viewDots stat ]
        ]

viewDots : Stat -> Html Msg
viewDots stat =
    let
        dotSize = 12*2
        dotGap = dotSize / 2
        dotOutline = 0.05 * dotSize
        marginalIncrease = (  dotSize + dotGap ) / 2
        svgOutlineMargin = ( sqrt ( 2 * ( dotOutline ^ 2 ) ) ) / 2

        rangeList = List.range 0 ( stat.dots.max - 1 )

        isFilledList = rangeList |> List.map ( (>) stat.dots.value )
    in
        svg 
            [ Svg.Attributes.width  ( String.fromFloat ( dotSize + marginalIncrease * toFloat ( stat.dots.max - 1 ) + 2 * svgOutlineMargin ) )
            , Svg.Attributes.height ( String.fromFloat ( dotSize + marginalIncrease + 2 * svgOutlineMargin ) )
            ]
            ( rangeList
                |> List.map ( toDotPosition dotSize dotGap svgOutlineMargin )
                |> List.map3 ( viewDot stat dotSize dotOutline ) isFilledList rangeList
            )

toDotPosition : Float -> Float -> Float -> Int -> ( Float, Float )
toDotPosition size gap offset id =
    let
        marginalIncrease = ( size + gap ) / 2
    in
        ( toFloat id * marginalIncrease + offset
        , toFloat ( remainderBy 2 ( id + 1 ) ) * marginalIncrease + offset
        )

viewDot : Stat -> Float -> Float -> Bool -> Int -> ( Float, Float ) -> Html Msg
viewDot stat size strokeWidth isFilled id position =
    Svg.polygon 
        [ Svg.Attributes.points 
            (   [ ( 0.5, 0.0 )
                , ( 1.0, 0.5 )
                , ( 0.5, 1.0 )
                , ( 0.0, 0.5 )
                ]
                |> List.map ( Tuple.mapBoth   ( (*) size ) ( (*) size ) )
                |> List.map ( Tuple.mapFirst  ( (+) ( Tuple.first  position ) ) )
                |> List.map ( Tuple.mapSecond ( (+) ( Tuple.second position ) ) )
                |> List.map tupleToString
                |> List.intersperse " "
                |> String.concat
            )
        , Svg.Attributes.fill "var(--bulma-body-color)"
        , Svg.Attributes.fillOpacity ( boolAsIntString isFilled )
        , Svg.Attributes.stroke "var(--bulma-body-color)"
        , Svg.Attributes.strokeWidth ( String.fromFloat strokeWidth )
        , onClick ( ChangeStat { stat | dots = BoundedInt.changeTo stat.dots ( valueOrReset stat.dots ( id + 1 ) ) } )
        ] []

tupleToString : ( Float, Float ) -> String
tupleToString tuple =
    let
        first  = String.fromFloat ( Tuple.first  tuple )
        second = String.fromFloat ( Tuple.second tuple )
    in
        first ++ "," ++ second

boolAsIntString : Bool -> String
boolAsIntString bool =
    if bool
    then "1"
    else "0"

valueOrReset : BoundedInt -> Int -> Int
valueOrReset dots value =
    if dots.value == value
    then dots.min
    else value

viewImport : Html Msg
viewImport =
    button 
        [ class "button"
        , onClick ImportCharacter
        ] [ text "Import Character" ]

viewModal : Model -> Html Msg
viewModal model =
    case model.modal of
        Nothing -> div [ class "modal" ] []
        Just modalType ->
            div [ class "modal is-active" ]
                [ div 
                    [ class "modal-background" 
                    , onClick CloseModal
                    ] []
                , div [ class "modal-card" ] 
                    [ header [ class "modal-card-head" ]
                        [ p [ class "modal-card-title" ] [ text ( "Change " ++ ( modalTitle modalType ) ) ]
                        ]
                    , section [ class "modal-card-body" ] 
                        [ p [] 
                            [ input 
                                [ class "input" 
                                , type_ "text"
                                , placeholder ( "Enter New " ++ ( modalTitle modalType ) )
                                , value model.modalValue
                                , onInput ( \x -> ChangeModalValue x )
                                ] []
                            ]
                        ]
                    , footer [ class "modal-card-foot" ] 
                        [ div [ class "buttons" ] 
                            [ button 
                                [ class "button is-success"
                                , onClick ( SaveModalChange modalType )
                                ] [ text "Save Changes" ]
                            , button 
                                [ class "button"
                                , onClick CloseModal
                                ] [ text "Cancel" ]
                            ]
                        ]
                    ]
                ]

modalTitle : ModalType -> String
modalTitle modalType =
    case modalType of
        Name -> "Character Name"

modalValue : ModalType -> Character -> String
modalValue modalType character =
    case modalType of
        Name -> character.name

subscriptions : Model -> Sub Msg
subscriptions _ =
  Sub.none

-- aria functions

ariaLabel : String -> Attribute msg
ariaLabel value =
    Html.Attributes.attribute "aria-label" value

ariaHaspopup : String -> Attribute msg
ariaHaspopup value =
    Html.Attributes.attribute "aria-haspopup" value

ariaControls : String -> Attribute msg
ariaControls value =
    Html.Attributes.attribute "aria-controls" value

ariaHidden : String -> Attribute msg
ariaHidden value =
    Html.Attributes.attribute "aria-hidden" value

ariaRole : String -> Attribute msg
ariaRole value =
    Html.Attributes.attribute "role" value

-- json decoder

jsonDecoder : Json.Decode.Decoder Character
jsonDecoder =
    Json.Decode.map3 Character
        ( Json.Decode.field   "name"  Json.Decode.string )
        ( Json.Decode.succeed         Creation.new       )
        ( Json.Decode.field   "stats" jsonStatsDecoder   )


jsonStatsDecoder : Json.Decode.Decoder ( List Stat )
jsonStatsDecoder =
    Json.Decode.map5 ( \spheres arete attributes abilities backgrounds -> spheres ++ ( List.singleton arete ) ++ attributes ++ abilities ++ backgrounds )
        ( Json.Decode.field "spheres"     <| Json.Decode.list jsonSphereDecoder     )
        ( Json.Decode.field "arete"          jsonAreteDecoder                       )
        ( Json.Decode.field "attributes"  <| jsonAttributesDecoder                  )
        ( Json.Decode.field "abilities"   <| jsonAbilitiesDecoder                   )
        ( Json.Decode.field "backgrounds" <| Json.Decode.list jsonBackgroundDecoder )

jsonSphereDecoder : Json.Decode.Decoder Stat
jsonSphereDecoder =
    Json.Decode.map3 ( \name dots affinity -> Stats.newSphere name |> Stats.changeDotsTo dots |> Stats.setAffinityNoChecks affinity )
        ( Json.Decode.field "name"     Json.Decode.string )
        ( Json.Decode.field "dots"     Json.Decode.int    )
        ( Json.Decode.field "affinity" Json.Decode.bool   )

jsonAreteDecoder : Json.Decode.Decoder Stat
jsonAreteDecoder =
    Json.Decode.map ( \dots -> Stats.newArete |> Stats.changeDotsTo dots )
        ( Json.Decode.field "dots" Json.Decode.int )

jsonAttributesDecoder : Json.Decode.Decoder ( List Stat )
jsonAttributesDecoder =
    Json.Decode.map3 ( \physical social mental -> physical ++ social ++ mental )
        ( Json.Decode.field "physical" <| Json.Decode.list ( jsonAttributeDecoder Stats.Physical ) )
        ( Json.Decode.field "social"   <| Json.Decode.list ( jsonAttributeDecoder Stats.Social   ) )
        ( Json.Decode.field "mental"   <| Json.Decode.list ( jsonAttributeDecoder Stats.Mental   ) )

jsonAttributeDecoder : Stats.Category -> Json.Decode.Decoder Stat
jsonAttributeDecoder cat =
    Json.Decode.map2 ( \name dots -> Stats.newAttribute cat name |> Stats.changeDotsTo dots )
        ( Json.Decode.field "name" Json.Decode.string )
        ( Json.Decode.field "dots" Json.Decode.int    )

jsonAbilitiesDecoder : Json.Decode.Decoder ( List Stat )
jsonAbilitiesDecoder =
    Json.Decode.map3 ( \talents skills knowledges -> talents ++ skills ++ knowledges )
        ( Json.Decode.field "talents"    <| Json.Decode.list ( jsonAbilityDecoder Stats.Skill     ) )
        ( Json.Decode.field "skills"     <| Json.Decode.list ( jsonAbilityDecoder Stats.Talent    ) )
        ( Json.Decode.field "knowledges" <| Json.Decode.list ( jsonAbilityDecoder Stats.Knowledge ) )

jsonAbilityDecoder : Stats.Category -> Json.Decode.Decoder Stat
jsonAbilityDecoder cat =
    Json.Decode.map2 ( \name dots -> Stats.newAbility cat name |> Stats.changeDotsTo dots )
        ( Json.Decode.field "name" Json.Decode.string )
        ( Json.Decode.field "dots" Json.Decode.int    )

jsonBackgroundDecoder : Json.Decode.Decoder Stat
jsonBackgroundDecoder =
    Json.Decode.map3 ( \name dots max -> Stats.newBackground name max |> Stats.changeDotsTo dots )
        ( Json.Decode.field "name" Json.Decode.string )
        ( Json.Decode.field "dots" Json.Decode.int    )
        ( Json.Decode.field "max"  Json.Decode.int    )